from __future__ import annotations

from pathlib import Path
import yaml

from filemindr.core.runner import run_pipeline


def write_cfg(path: Path, cfg: dict) -> None:
    path.write_text(yaml.safe_dump(cfg, sort_keys=False), encoding="utf-8")


def test_rule_conflict_policy_overrides_global_rename(tmp_path: Path):
    source = tmp_path / "Downloads"
    images = tmp_path / "images"
    source.mkdir()
    images.mkdir()

    (source / "pic.jpg").write_text("NEW", encoding="utf-8")

    (images / "pic.jpg").write_text("OLD", encoding="utf-8")

    cfg = {
        "source": str(source),
        "default_target": str(source / "others"),
        "conflict_policy": "rename",
        "rules": [
            {
                "name": "images",
                "priority": 10,
                "match": {"extensions": ["jpg"]},
                "action": {
                    "move_to": str(images),
                    "conflict_policy": "overwrite",
                },
            }
        ],
    }

    cfg_path = tmp_path / "filemindr.yaml"
    write_cfg(cfg_path, cfg)

    run_pipeline(str(cfg_path), dry_run=False)

    assert (images / "pic.jpg").read_text(encoding="utf-8") == "NEW"
    assert not (source / "pic.jpg").exists()


def test_rule_without_policy_uses_global(tmp_path: Path):
    source = tmp_path / "Downloads"
    docs = tmp_path / "docs"
    source.mkdir()
    docs.mkdir()

    (source / "a.txt").write_text("NEW", encoding="utf-8")
    (docs / "a.txt").write_text("OLD", encoding="utf-8")

    cfg = {
        "source": str(source),
        "default_target": str(source / "others"),
        "conflict_policy": "rename",
        "rules": [
            {
                "name": "docs",
                "priority": 10,
                "match": {"extensions": ["txt"]},
                "action": {
                    "move_to": str(docs),
                },
            }
        ],
    }

    cfg_path = tmp_path / "filemindr.yaml"
    write_cfg(cfg_path, cfg)

    run_pipeline(str(cfg_path), dry_run=False)

    assert (docs / "a.txt").read_text(encoding="utf-8") == "OLD"
    assert (docs / "a (1).txt").read_text(encoding="utf-8") == "NEW"
    assert not (source / "a.txt").exists()


from __future__ import annotations

from pathlib import Path

import pytest
from typer.testing import CliRunner

from filemindr.cli import app
from filemindr.core.config import resolve_profile_config

runner = CliRunner()


def _fake_home(monkeypatch: pytest.MonkeyPatch, fake_home: Path) -> None:
    fake_home.mkdir(parents=True, exist_ok=True)
    monkeypatch.setattr(Path, "home", lambda: fake_home)


def _profile_paths(fake_home: Path, profile: str) -> tuple[Path, Path, Path]:
    """
    Returns:
      base_dir, rules_yaml_path, profiles_yaml_path
    """
    base = fake_home / ".filemindr"
    rules_yaml = base / "rules" / profile / "rules.yaml"
    profiles_yaml = base / "profiles.yaml"
    return base, rules_yaml, profiles_yaml


def test_profile_init_creates_rules_and_registry(fake_home: Path, runner: CliRunner):
    result = runner.invoke(app, ["profile", "init", "home"])
    assert result.exit_code == 0, (result.stdout or "") + (result.stderr or "")

    _, rules_yaml, profiles_yaml = _profile_paths(fake_home, "home")

    assert rules_yaml.exists()
    assert profiles_yaml.exists()

    text = rules_yaml.read_text(encoding="utf-8")
    assert "filemindr configuration" in text
    assert "home" in profiles_yaml.read_text(encoding="utf-8")


def test_profile_init_refuses_overwrite_without_force(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    fake_home = tmp_path / "home"
    _fake_home(monkeypatch, fake_home)

    r1 = runner.invoke(app, ["profile", "init", "home"])
    assert r1.exit_code == 0

    r2 = runner.invoke(app, ["profile", "init", "home"])
    assert r2.exit_code != 0


def test_profile_init_force_overwrites_rules_yaml(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    fake_home = tmp_path / "home"
    _fake_home(monkeypatch, fake_home)

    r1 = runner.invoke(app, ["profile", "init", "home"])
    assert r1.exit_code == 0

    _, rules_yaml, _ = _profile_paths(fake_home, "home")

    rules_yaml.write_text("old", encoding="utf-8")

    r2 = runner.invoke(app, ["profile", "init", "home", "--force"])
    assert r2.exit_code == 0

    text = rules_yaml.read_text(encoding="utf-8")
    assert text != "old"
    assert "filemindr configuration" in text


def test_resolve_profile_config_returns_rules_yaml(fake_home: Path, runner: CliRunner):
    runner.invoke(app, ["profile", "init", "home"])
    resolved = Path(resolve_profile_config("home"))
    assert resolved.exists()
    assert resolved.name == "rules.yaml"

def test_resolve_profile_config_raises_when_profile_missing(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    fake_home = tmp_path / "home"
    _fake_home(monkeypatch, fake_home)

    with pytest.raises(Exception):
        resolve_profile_config("missing-profile")



from __future__ import annotations

import re
from dataclasses import dataclass
from pathlib import Path

import yaml

ALLOWED_POLICIES = {"rename", "skip", "overwrite", "trash"}


@dataclass(frozen=True)
class ValidationResult:
    ok: bool
    errors: list[str]


def validate_config_file(config_path: Path) -> ValidationResult:
    errors: list[str] = []

    if not config_path.exists():
        return ValidationResult(ok=False, errors=[f"Config not found: {config_path}"])

    try:
        raw = config_path.read_text(encoding="utf-8")
    except Exception as e:
        return ValidationResult(ok=False, errors=[f"Failed to read config as UTF-8: {e}"])

    try:
        config = yaml.safe_load(raw) or {}
    except Exception as e:
        return ValidationResult(ok=False, errors=[f"Invalid YAML: {e}"])

    source_raw = config.get("source")
    if not source_raw or not isinstance(source_raw, str):
        errors.append("Missing required field: source (string)")
        return ValidationResult(ok=False, errors=errors)

    source_dir = Path(source_raw).expanduser()
    if not source_dir.exists():
        errors.append(f"source directory does not exist: {source_dir}")

    global_policy = str(config.get("conflict_policy", "rename")).lower()
    if global_policy not in ALLOWED_POLICIES:
        errors.append(
            f"Invalid global conflict_policy='{global_policy}'. "
            f"Allowed: {', '.join(sorted(ALLOWED_POLICIES))}"
        )

    rules_cfg = config.get("rules", [])
    if rules_cfg is None:
        rules_cfg = []

    if not isinstance(rules_cfg, list):
        errors.append("Field 'rules' must be a list")
        return ValidationResult(ok=False, errors=errors)

    for i, r in enumerate(rules_cfg):
        if not isinstance(r, dict):
            errors.append(f"rules[{i}] must be an object")
            continue

        name = r.get("name") or f"rules[{i}]"
        if not isinstance(name, str):
            name = f"rules[{i}]"

        prio = r.get("priority", 0)
        try:
            int(prio)
        except Exception:
            errors.append(f"Rule '{name}': priority must be an integer")

        match = r.get("match", {}) or {}
        if not isinstance(match, dict):
            errors.append(f"Rule '{name}': match must be an object")
            match = {}

        regex_raw = match.get("regex")
        if regex_raw is not None:
            if not isinstance(regex_raw, str):
                errors.append(f"Rule '{name}': match.regex must be a string")
            else:
                try:
                    re.compile(regex_raw)
                except Exception as e:
                    errors.append(f"Rule '{name}': invalid regex '{regex_raw}': {e}")

        older = match.get("older_than_days")
        if older is not None:
            try:
                older_i = int(older)
                if older_i < 0:
                    errors.append(f"Rule '{name}': older_than_days must be >= 0")
            except Exception:
                errors.append(f"Rule '{name}': older_than_days must be an integer")

        exts = match.get("extensions", [])
        if exts is not None:
            if not isinstance(exts, list):
                errors.append(f"Rule '{name}': match.extensions must be a list")
            else:
                for e in exts:
                    if not isinstance(e, str):
                        errors.append(f"Rule '{name}': extension must be a string (got {type(e).__name__})")
                        break

        action = r.get("action", {}) or {}
        if not isinstance(action, dict):
            errors.append(f"Rule '{name}': action must be an object")
            action = {}

        move_to = action.get("move_to")
        copy_to = action.get("copy_to")

        if bool(move_to) == bool(copy_to):
            errors.append(
                f"Rule '{name}': must define exactly one of action.move_to or action.copy_to"
            )

        rule_policy = action.get("conflict_policy")
        if rule_policy is not None:
            rp = str(rule_policy).lower()
            if rp not in ALLOWED_POLICIES:
                errors.append(
                    f"Rule '{name}': invalid conflict_policy='{rp}'. "
                    f"Allowed: {', '.join(sorted(ALLOWED_POLICIES))}"
                )

    return ValidationResult(ok=(len(errors) == 0), errors=errors)



from __future__ import annotations

import re
import shutil
import time
from collections import Counter
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import yaml
from loguru import logger


def _p(value: str) -> Path:
    return Path(value).expanduser()


@dataclass(frozen=True)
class Rule:
    name: str
    priority: int
    extensions: set[str]
    regex: re.Pattern | None
    older_than_days: int | None
    move_to: Path | None
    copy_to: Path | None
    conflict_policy: str | None

def _trash(path: Path) -> None:
    """
    Send a file to the OS trash/recycle bin.
    Falls back to unlink if trash is not available.
    """
    try:
        from send2trash import send2trash
        send2trash(str(path))
    except Exception:
        path.unlink(missing_ok=True)


def _normalize_ext(ext: str) -> str:
    return ext.lower().lstrip(".")


def _load_rules(config: dict[str, Any]) -> list[Rule]:
    rules_cfg = config.get("rules", [])
    rules: list[Rule] = []

    for r in rules_cfg:
        name = r.get("name", "unnamed")
        priority = int(r.get("priority", 0))

        match = r.get("match", {}) or {}
        exts = {_normalize_ext(e) for e in match.get("extensions", [])}
        regex_raw = match.get("regex")
        regex = re.compile(regex_raw) if regex_raw else None
        older = match.get("older_than_days")
        older_than_days = int(older) if older is not None else None

        action = r.get("action", {}) or {}
        move_to_str = action.get("move_to")
        copy_to_str = action.get("copy_to")

        rule_policy = action.get("conflict_policy")

        if bool(move_to_str) == bool(copy_to_str):
            raise ValueError(
                f"Rule '{name}' must define exactly one of action.move_to or action.copy_to"
            )

        rules.append(
            Rule(
                name=name,
                priority=priority,
                extensions=exts,
                regex=regex,
                older_than_days=older_than_days,
                move_to=_p(move_to_str) if move_to_str else None,
                copy_to=_p(copy_to_str) if copy_to_str else None,
                conflict_policy=str(rule_policy) if rule_policy else None,
            )
        )

    rules.sort(key=lambda x: x.priority, reverse=True)
    return rules


def _is_older_than(file: Path, days: int) -> bool:
    cutoff_seconds = days * 24 * 60 * 60
    age_seconds = time.time() - file.stat().st_mtime
    return age_seconds >= cutoff_seconds


def _match_rule(file: Path, rules: list[Rule]) -> Rule | None:
    ext = _normalize_ext(file.suffix)
    filename = file.name

    for rule in rules:
        if rule.extensions and ext not in rule.extensions:
            continue

        if rule.regex and not rule.regex.search(filename):
            continue

        if rule.older_than_days is not None:
            if not _is_older_than(file, rule.older_than_days):
                continue

        return rule

    return None

def _resolve_conflict(dest: Path, policy: str) -> Path | None:
    """
    Returns:
      - Path to use (possibly renamed), or
      - None if policy == skip and file exists.
    """
    if not dest.exists():
        return dest

    policy = policy.lower()

    if policy in {"overwrite", "trash"}:
        return dest

    if policy == "skip":
        return None

    stem, suffix = dest.stem, dest.suffix
    parent = dest.parent
    i = 1
    while True:
        candidate = parent / f"{stem} ({i}){suffix}"
        if not candidate.exists():
            return candidate
        i += 1


def run_pipeline(config_path: str, dry_run: bool = False, only_paths: set[Path] | None = None) -> None:
    cfg_path = Path(config_path)
    if not cfg_path.exists():
        raise FileNotFoundError(config_path)

    config = yaml.safe_load(cfg_path.read_text()) or {}
    cfg_abs = cfg_path.expanduser().resolve()

    source = _p(config["source"])
    default_target = _p(config.get("default_target", str(source / "others")))
    global_policy = str(config.get("conflict_policy", "rename"))

    rules = _load_rules(config)

    total_files = 0
    moved = 0
    skipped = 0
    overwritten = 0
    errors = 0

    by_rule = Counter()
    by_action = Counter()

    logger.info(f"Scanning: {source}")

    targets = {default_target} | {p for r in rules for p in (r.move_to, r.copy_to) if p}
    for t in targets:
        if dry_run:
            logger.debug(f"[DRY] ensure dir: {t}")
        else:
            t.mkdir(parents=True, exist_ok=True)

    for file in source.iterdir():
        if not file.is_file():
            continue
        if file.resolve() == cfg_abs:
            continue
        if only_paths is not None and file not in only_paths:
            continue

        total_files += 1

        rule = _match_rule(file, rules)
        rule_name = rule.name if rule else "default"
        dest_dir = (
            (rule.copy_to or rule.move_to) if rule else default_target
        )
        dest = dest_dir / file.name

        policy = (rule.conflict_policy if rule and rule.conflict_policy else global_policy)
        resolved = _resolve_conflict(dest, policy)
        if resolved is None:
            skipped += 1
            by_rule[rule_name] += 1
            by_action["skipped"] += 1
            logger.debug(f"SKIP (exists): {dest}")
            continue

        policy = policy.lower()
        will_replace = resolved.exists() and policy.lower() in {"overwrite", "trash"}

        if dry_run:
            by_rule[rule_name] += 1
            by_action["planned"] += 1
            chosen = f"rule={rule_name} prio={rule.priority}" if rule else "rule=default"
            action_name = "COPY" if (rule and rule.copy_to) else "MOVE"
            logger.debug(f"[DRY] {action_name} {chosen} | {file} -> {resolved}")
            continue

        try:
            if will_replace:
                if policy.lower() == "trash":
                    _trash(resolved)
                    by_action["trashed"] += 1
                else:
                    resolved.unlink()
                    overwritten += 1
                    by_action["overwritten"] += 1

            if rule and rule.copy_to:
                shutil.copy2(file, resolved)
            else:
                file.rename(resolved)

            moved += 1
            by_rule[rule_name] += 1
            by_action["moved"] += 1

            chosen = f"rule={rule_name} prio={rule.priority}" if rule else "rule=default"
            action_name = "COPY" if (rule and rule.copy_to) else "MOVE"
            logger.debug(f"{action_name} {chosen} | {file} -> {resolved}")
        except Exception as e:
            errors += 1
            by_action["errors"] += 1
            logger.exception(f"ERROR moving {file} -> {resolved}: {e}")


    logger.info("==== SUMMARY ====")
    logger.info(f"Dry-run: {dry_run}")
    logger.info(f"Files scanned: {total_files}")

    if dry_run:
        logger.info(f"Planned moves: {by_action.get('planned', 0)}")
    else:
        logger.info(f"Moved: {moved}")
        logger.info(f"Overwritten: {overwritten}")
        logger.info(f"Skipped: {skipped}")
        logger.info(f"Errors: {errors}")
        logger.info(f"Trashed: {by_action.get('trashed', 0)}")

    logger.info("By rule:")
    for rule_name, count in by_rule.most_common():
        logger.info(f"  - {rule_name}: {count}")


from __future__ import annotations

from pathlib import Path

import pytest
import typer
from typer.testing import CliRunner

from filemindr.cli import app


@pytest.fixture()
def runner() -> CliRunner:
    return CliRunner()


@pytest.fixture()
def fake_home(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> Path:
    fake_home = tmp_path / "home"
    fake_home.mkdir(parents=True, exist_ok=True)

    monkeypatch.setattr(Path, "home", lambda: fake_home)

    monkeypatch.setenv("HOME", str(fake_home))
    monkeypatch.setenv("USERPROFILE", str(fake_home))

    monkeypatch.setenv("HOMEDRIVE", fake_home.drive or "C:")

    homedir = str(fake_home)
    if ":" in homedir:
        homedir = homedir.split(":", 1)[1]
    homedir = homedir.replace("/", "\\")
    if not homedir.startswith("\\"):
        homedir = "\\" + homedir
    monkeypatch.setenv("HOMEPATH", homedir)

    return fake_home


@pytest.fixture()
def cli_app() -> typer.Typer:
    return app



from __future__ import annotations


def test_profile_remove_deletes_registry_and_folder(runner, fake_home, cli_app):
    runner.invoke(cli_app, ["profile", "init", "home"])

    base = fake_home / ".filemindr"
    profile_dir = base / "rules" / "home"
    assert profile_dir.exists()

    result = runner.invoke(cli_app, ["profile", "remove", "home", "--yes"])
    assert result.exit_code == 0

    assert not profile_dir.exists()

    profiles_file = base / "profiles.yaml"
    if profiles_file.exists():
        text = profiles_file.read_text(encoding="utf-8")
        assert "home" not in text



from __future__ import annotations


def test_profile_init_existing_without_force_fails(runner, fake_home, cli_app):
    runner.invoke(cli_app, ["profile", "init", "home"])

    result = runner.invoke(cli_app, ["profile", "init", "home"])
    assert result.exit_code != 0
    assert "already exists" in result.stdout.lower() or result.stderr.lower()



from __future__ import annotations


def test_profile_list_shows_profiles(runner, fake_home, cli_app):
    runner.invoke(cli_app, ["profile", "init", "home"])

    result = runner.invoke(cli_app, ["profile", "list"])
    assert result.exit_code == 0
    assert "home" in result.stdout



from pathlib import Path
import os
import yaml


def _expand(p):
    return Path(os.path.expandvars(p)).expanduser().resolve()


def resolve_profile_config(profile: str) -> Path:
    base = _expand("~/.filemindr")
    profiles_file = base / "profiles.yaml"

    if not profiles_file.exists():
        raise FileNotFoundError("profiles.yaml not found in ~/.filemindr | Run: filemindr profile init <name>")

    data = yaml.safe_load(profiles_file.read_text()) or {}
    profiles = data.get("profiles", {})

    if profile not in profiles:
        raise ValueError(f"Profile '{profile}' not defined")

    rules_dir = _expand(profiles[profile])
    rules_dir.mkdir(parents=True, exist_ok=True)

    rules_yaml = rules_dir / "rules.yaml"

    if not rules_yaml.exists():
        raise FileNotFoundError(rules_yaml)

    return rules_yaml



from __future__ import annotations

from pathlib import Path

import pytest
from typer.testing import CliRunner

from filemindr.cli import app

runner = CliRunner()


def _fake_home(monkeypatch: pytest.MonkeyPatch, fake_home: Path) -> None:
    fake_home.mkdir(parents=True, exist_ok=True)


    monkeypatch.setattr(Path, "home", lambda: fake_home)

    monkeypatch.setenv("HOME", str(fake_home))
    monkeypatch.setenv("USERPROFILE", str(fake_home))

    monkeypatch.setenv("HOMEDRIVE", fake_home.drive or "C:")
    monkeypatch.setenv(
        "HOMEPATH",
        "\\" + "\\".join(fake_home.parts[1:]) if fake_home.drive else str(fake_home).replace("/", "\\"),
    )


def _write_profile(fake_home: Path, profile: str, rules_yaml_text: str) -> None:
    base = fake_home / ".filemindr"
    rules_dir = base / "rules" / profile
    rules_dir.mkdir(parents=True, exist_ok=True)

    (rules_dir / "rules.yaml").write_text(rules_yaml_text, encoding="utf-8")

    (base / "profiles.yaml").write_text(
        f"profiles:\n  {profile}: {rules_dir.as_posix()}\n",
        encoding="utf-8",
    )


def test_validate_ok(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    fake_home = tmp_path / "home"
    _fake_home(monkeypatch, fake_home)

    src = tmp_path / "Downloads"
    src.mkdir()
    src_s = src.as_posix()

    rules = f"""
source: {src_s}
default_target: {src_s}/others
conflict_policy: rename

rules:
  - name: images
    priority: 10
    match:
      extensions: ["jpg"]
    action:
      move_to: {src_s}/images
""".strip()

    _write_profile(fake_home, "home", rules)

    result = runner.invoke(app, ["validate", "-p", "home"])
    assert result.exit_code == 0

    out = (result.stdout or "") + (result.stderr or "")
    assert "Config is valid" in out


def test_validate_invalid_regex(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    fake_home = tmp_path / "home"
    _fake_home(monkeypatch, fake_home)

    src = tmp_path / "Downloads"
    src.mkdir()
    src_s = src.as_posix()

    rules = f"""
source: {src_s}
default_target: {src_s}/others

rules:
  - name: bad
    priority: 1
    match:
      regex: "("
      extensions: ["pdf"]
    action:
      move_to: {src_s}/documents
""".strip()

    _write_profile(fake_home, "home", rules)

    result = runner.invoke(app, ["validate", "-p", "home"])
    assert result.exit_code == 1

    out = ((result.stdout or "") + (result.stderr or "")).lower()
    assert "regex" in out


def test_validate_rejects_move_and_copy_together(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    fake_home = tmp_path / "home"
    _fake_home(monkeypatch, fake_home)

    src = tmp_path / "Downloads"
    src.mkdir()
    src_s = src.as_posix()

    rules = f"""
source: {src_s}

rules:
  - name: both
    priority: 1
    match:
      extensions: ["pdf"]
    action:
      move_to: {src_s}/documents
      copy_to: {src_s}/backup
""".strip()

    _write_profile(fake_home, "home", rules)

    result = runner.invoke(app, ["validate", "-p", "home"])
    assert result.exit_code == 1

    out = ((result.stdout or "") + (result.stderr or "")).lower()
    assert "move_to" in out and "copy_to" in out



from __future__ import annotations

from pathlib import Path

import pytest


from filemindr.core.runner import run_pipeline


def _write_yaml(path: Path, content: str) -> None:
    path.write_text(content, encoding="utf-8")


def test_copy_to_keeps_source_and_copies_to_target(tmp_path: Path) -> None:
    source = tmp_path / "Downloads"
    target = tmp_path / "images"
    source.mkdir(parents=True)
    target.mkdir(parents=True)

    src_file = source / "photo.png"
    src_file.write_bytes(b"fake-png")

    config_path = tmp_path / "filemindr.yaml"
    _write_yaml(
        config_path,
        f"""
source: "{source.as_posix()}"
default_target: "{(tmp_path / "others").as_posix()}"
conflict_policy: rename

rules:
  - name: images_copy
    priority: 10
    match:
      extensions: ["png"]
    action:
      copy_to: "{target.as_posix()}"
""".lstrip(),
    )

    run_pipeline(str(config_path), dry_run=False)

    assert src_file.exists()

    copied = target / "photo.png"
    assert copied.exists()
    assert copied.read_bytes() == b"fake-png"


def test_move_to_removes_source_and_moves_to_target(tmp_path: Path) -> None:
    source = tmp_path / "Downloads"
    target = tmp_path / "documents"
    source.mkdir(parents=True)
    target.mkdir(parents=True)

    src_file = source / "doc.txt"
    src_file.write_text("hello", encoding="utf-8")

    config_path = tmp_path / "filemindr.yaml"
    _write_yaml(
        config_path,
        f"""
source: "{source.as_posix()}"
default_target: "{(tmp_path / "others").as_posix()}"
conflict_policy: rename

rules:
  - name: docs_move
    priority: 10
    match:
      extensions: ["txt"]
    action:
      move_to: "{target.as_posix()}"
""".lstrip(),
    )

    run_pipeline(str(config_path), dry_run=False)

    assert not src_file.exists()

    moved = target / "doc.txt"
    assert moved.exists()
    assert moved.read_text(encoding="utf-8") == "hello"


def test_rule_cannot_have_move_to_and_copy_to(tmp_path: Path) -> None:
    source = tmp_path / "Downloads"
    source.mkdir(parents=True)

    config_path = tmp_path / "filemindr.yaml"
    _write_yaml(
        config_path,
        f"""
source: "{source.as_posix()}"
default_target: "{(tmp_path / "others").as_posix()}"
conflict_policy: rename

rules:
  - name: invalid_both
    priority: 10
    match:
      extensions: ["txt"]
    action:
      move_to: "{(tmp_path / "docs").as_posix()}"
      copy_to: "{(tmp_path / "docs_copy").as_posix()}"
""".lstrip(),
    )

    with pytest.raises(ValueError, match="must define exactly one"):
        run_pipeline(str(config_path), dry_run=False)


def test_rule_must_have_move_to_or_copy_to(tmp_path: Path) -> None:
    source = tmp_path / "Downloads"
    source.mkdir(parents=True)

    config_path = tmp_path / "filemindr.yaml"
    _write_yaml(
        config_path,
        f"""
source: "{source.as_posix()}"
default_target: "{(tmp_path / "others").as_posix()}"
conflict_policy: rename

rules:
  - name: invalid_none
    priority: 10
    match:
      extensions: ["txt"]
    action:
      # nenhum move_to/copy_to
      something_else: "x"
""".lstrip(),
    )

    with pytest.raises(ValueError, match="must define exactly one"):
        run_pipeline(str(config_path), dry_run=False)


from __future__ import annotations

import os
from pathlib import Path

from typer.testing import CliRunner

from filemindr.cli import app

runner = CliRunner()


def _write_profile(fake_home: Path, profile: str, rules_yaml: str) -> None:
    """
    Create:
      ~/.filemindr/profiles.yaml
      ~/.filemindr/rules/<profile>/rules.yaml
    under fake_home (which is Path.home()).
    """
    base = fake_home / ".filemindr"
    rules_dir = base / "rules" / profile
    rules_dir.mkdir(parents=True, exist_ok=True)

    (rules_dir / "rules.yaml").write_text(rules_yaml, encoding="utf-8")

    profiles_yaml = base / "profiles.yaml"
    profiles_yaml.write_text(
        f"profiles:\n  {profile}: {rules_dir.as_posix()}\n",
        encoding="utf-8",
    )


def _rules_for(source_dir: Path) -> str:
    src = source_dir.as_posix()
    return f"""
source: {src}
default_target: {src}/others
conflict_policy: rename

rules:
  - name: images
    priority: 50
    match:
      extensions: ["jpg", "png"]
    action:
      move_to: {src}/images
""".lstrip()


def test_explain_marks_directories_as_skip(tmp_path: Path, fake_home: Path, runner: CliRunner):
    _write_profile(fake_home, "home", _rules_for(tmp_path))

    (tmp_path / "documents").mkdir()
    (tmp_path / "pic.jpg").write_text("x", encoding="utf-8")

    result = runner.invoke(app, ["explain", str(tmp_path), "-p", "home"])
    assert result.exit_code == 0, (result.stdout or "") + (result.stderr or "")

def test_explain_respects_default_limit(tmp_path: Path, fake_home: Path):
    _write_profile(fake_home, "home", _rules_for(tmp_path))

    for i in range(0, 80):
        (tmp_path / f"f{i:03d}.txt").write_text("x", encoding="utf-8")

    result = runner.invoke(app, ["explain", str(tmp_path), "-p", "home"])
    assert result.exit_code == 0

    out = (result.stdout or "") + (result.stderr or "")

    assert "Showing first" in out

    assert "f000.txt" in out
    assert "f049.txt" in out
    assert "f079.txt" not in out


def test_explain_limit_option_overrides(tmp_path: Path, fake_home: Path, runner: CliRunner):
    _write_profile(fake_home, "home", _rules_for(tmp_path))

    for i in range(0, 30):
        (tmp_path / f"f{i:03d}.txt").write_text("x", encoding="utf-8")

    result = runner.invoke(app, ["explain", str(tmp_path), "-p", "home", "--limit", "10"])
    assert result.exit_code == 0, (result.stdout or "") + (result.stderr or "")

    out = (result.stdout or "") + (result.stderr or "")
    assert "f000.txt" in out
    assert "f009.txt" in out
    assert "f010.txt" not in out


def test_explain_all_disables_limit(tmp_path: Path, fake_home: Path, runner: CliRunner):
    _write_profile(fake_home, "home", _rules_for(tmp_path))

    for i in range(0, 60):
        (tmp_path / f"f{i:03d}.txt").write_text("x", encoding="utf-8")

    result = runner.invoke(app, ["explain", str(tmp_path), "-p", "home", "--all"])
    assert result.exit_code == 0, (result.stdout or "") + (result.stderr or "")

    out = (result.stdout or "") + (result.stderr or "")
    assert "f000.txt" in out
    assert "f059.txt" in out
    assert "Showing first" not in out



DEFAULT_CONFIG = """\
# filemindr configuration (profile)
# Generated by `filemindr profile init <name>`
#
# Quick start:
#   filemindr validate -p <name>
#   filemindr run -p <name> --dry-run
#   filemindr run -p <name>
#   filemindr watch -p <name> --once
#
# Notes:
# - source: directory to scan (and watch).
# - default_target: where unmatched files go.
# - conflict_policy: what to do if destination file already exists:
#     rename | skip | overwrite | trash

source: ~/Downloads

# Anything that doesn't match a rule goes here
default_target: ~/Downloads/others

# Default conflict policy applied to all rules unless overridden per-rule
conflict_policy: rename  # rename | skip | overwrite | trash

rules:
  - name: invoices
    priority: 100
    match:
      regex: "(?i)invoice|receipt|irpf|recibo|nota"
      extensions: ["pdf"]
    action:
      move_to: ~/Downloads/finance/invoices

  - name: old_installers
    priority: 80
    match:
      extensions: ["exe", "msi"]
      older_than_days: 14
    action:
      move_to: ~/Downloads/installers/old

  - name: archives
    priority: 70
    match:
      extensions: ["zip", "rar", "7z", "tar", "gz"]
    action:
      move_to: ~/Downloads/archives

  - name: videos
    priority: 60
    match:
      extensions: ["mp4", "mkv", "mov", "avi", "webm", "m4v"]
    action:
      move_to: ~/Downloads/videos

  - name: audio
    priority: 55
    match:
      extensions: ["mp3", "wav", "m4a", "aac", "flac", "ogg"]
    action:
      move_to: ~/Downloads/audio

  - name: images
    priority: 50
    match:
      extensions: ["jpg", "jpeg", "png", "webp", "gif", "svg"]
    action:
      move_to: ~/Downloads/images

  - name: documents
    priority: 40
    match:
      extensions: ["pdf", "docx", "xlsx", "pptx", "txt", "md"]
    action:
      move_to: ~/Downloads/documents

  - name: code_snippets
    priority: 30
    match:
      extensions: ["py", "js", "ts", "json", "yaml", "yml", "toml"]
    action:
      move_to: ~/Downloads/code
"""



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any

import yaml

from filemindr.core.runner import (
    _p,
    _load_rules,
    _match_rule,
    _resolve_conflict,
)


@dataclass(frozen=True)
class ExplainResult:
    file: Path
    action: str
    rule_name: str
    rule_priority: int | None
    policy: str
    dest: Path
    resolved: Path | None
    reason: str | None


def _load_config(config_path: str) -> dict[str, Any]:
    cfg_path = Path(config_path)
    if not cfg_path.exists():
        raise FileNotFoundError(config_path)
    return yaml.safe_load(cfg_path.read_text(encoding="utf-8")) or {}


def explain_files(
    config_path: str,
    files: list[Path],
    *,
    verbose: bool = False,
) -> list[ExplainResult]:
    """
    Explain what would happen to each file. Does NOT move/copy anything.
    """
    config = _load_config(config_path)

    source = _p(config["source"]).resolve()
    default_target = _p(config.get("default_target", str(source / "others"))).resolve()
    global_policy = str(config.get("conflict_policy", "rename")).lower()

    rules = _load_rules(config)

    results: list[ExplainResult] = []

    for f in files:
        f = f.expanduser().resolve()

        if not f.exists() or not f.is_file():
            results.append(
                ExplainResult(
                    file=f,
                    action="SKIP",
                    rule_name="n/a",
                    rule_priority=None,
                    policy=global_policy,
                    dest=f,
                    resolved=None,
                    reason="not_found_or_not_file" if verbose else None,
                )
            )
            continue

        rule = _match_rule(f, rules)

        rule_name = rule.name if rule else "default"
        rule_priority = rule.priority if rule else None

        dest_dir = (rule.copy_to or rule.move_to) if rule else default_target
        dest_dir = dest_dir.resolve()

        action = "COPY" if (rule and rule.copy_to) else "MOVE"

        policy = (rule.conflict_policy if rule and rule.conflict_policy else global_policy).lower()

        dest = dest_dir / f.name
        resolved = _resolve_conflict(dest, policy)

        reason = None
        if verbose:
            parts: list[str] = []
            parts.append(f"source={source}")
            parts.append(f"ext={f.suffix.lower() or '-'}")

            if rule:
                parts.append(f"matched_rule={rule.name}")
                if rule.regex:
                    parts.append("regex=YES")
                if rule.older_than_days is not None:
                    parts.append(f"older_than_days={rule.older_than_days}")
            else:
                parts.append("matched_rule=default")

            if dest.exists():
                parts.append("dest_exists=YES")
            else:
                parts.append("dest_exists=NO")

            if resolved is None:
                parts.append("conflict=skip")
            elif resolved != dest:
                parts.append(f"conflict=rename -> {resolved.name}")
            else:
                if policy in {"overwrite", "trash"} and dest.exists():
                    parts.append(f"conflict={policy}")
                else:
                    parts.append("conflict=none")

            reason = " ".join(parts)

        results.append(
            ExplainResult(
                file=f,
                action=action,
                rule_name=rule_name,
                rule_priority=rule_priority,
                policy=policy,
                dest=dest,
                resolved=resolved,
                reason=reason,
            )
        )

    return results


def format_explain(result: ExplainResult, *, fmt: str = "one", verbose: bool = False) -> str:
    """
    fmt:
      - one:  <FILE> -> <DEST> [ACTION] rule=... prio=... policy=... (optional reason)
      - short: ACTION <FILE> -> <DEST> (rule=...)
    """
    file_name = result.file.name

    if result.action == "SKIP":
        base = f"{file_name} [SKIP]"
        if verbose and result.reason:
            base += f" ({result.reason})"
        return base

    if result.resolved is None:
        dest_str = str(result.dest)
        base = f"{file_name} -> {dest_str} [SKIP] rule={result.rule_name} prio={result.rule_priority or '-'} policy={result.policy}"
        if verbose and result.reason:
            base += f" ({result.reason})"
        return base

    dest_str = str(result.resolved)

    if fmt == "short":
        base = f"{result.action} {file_name} -> {dest_str} (rule={result.rule_name})"
        if verbose and result.reason:
            base += f" ({result.reason})"
        return base

    base = (
        f"{file_name} -> {dest_str} "
        f"[{result.action}] rule={result.rule_name} prio={result.rule_priority or '-'} policy={result.policy}"
    )
    if verbose and result.reason:
        base += f" ({result.reason})"
    return base



import sys
from pathlib import Path

import typer
import yaml
from loguru import logger

from filemindr.core.config import resolve_profile_config, _expand
from filemindr.core.explain import explain_files, format_explain
from filemindr.core.helpers import open_in_editor, open_with_default_app
from filemindr.core.runner import run_pipeline
from filemindr.core.templates.yaml_tmpl import DEFAULT_CONFIG
from filemindr.core.validator import validate_config_file
from filemindr.core.watcher import WatchOptions, watch_and_run


app = typer.Typer(help="Declarative local file pipelines")
profile_app = typer.Typer(help="Manage profiles")
app.add_typer(profile_app, name="profile")


def _setup_logger(level: str) -> None:
    logger.remove()
    logger.add(sys.stdout, level=level.upper())


def _home() -> Path:
    return Path.home() / ".filemindr"


def _profiles_file() -> Path:
    return _home() / "profiles.yaml"


@app.command()
def run(
    profile: str = typer.Option(..., "--profile", "-p"),
    dry_run: bool = False,
    log_level: str = typer.Option("INFO"),
):
    _setup_logger(log_level)

    config_path = resolve_profile_config(profile)
    logger.info(f"Running pipeline | profile={profile} config={config_path} dry_run={dry_run}")

    run_pipeline(str(config_path), dry_run)


@app.command()
def watch(
    profile: str = typer.Option(..., "--profile", "-p"),
    dry_run: bool = False,
    debounce_ms: int = 500,
    stable_ms: int = 1500,
    once: bool = typer.Option(False, "--once"),
    log_level: str = typer.Option("INFO"),
):
    _setup_logger(log_level)

    config_path = resolve_profile_config(profile)
    logger.info(f"Starting watch | profile={profile} config={config_path} dry_run={dry_run}")

    cfg = yaml.safe_load(Path(config_path).read_text(encoding="utf-8")) or {}
    source_dir = Path(cfg["source"]).expanduser()

    opts = WatchOptions(debounce_ms=debounce_ms, stable_ms=stable_ms)

    watch_and_run(
        source_dir=source_dir,
        config_path=str(config_path),
        dry_run=dry_run,
        opts=opts,
        once=once,
    )


@app.command()
def explain(
    paths: list[Path] = typer.Argument(...),
    profile: str = typer.Option(..., "--profile", "-p"),
    fmt: str = typer.Option("one", "--format", "-f"),
    verbose: bool = typer.Option(False, "--verbose", "-v"),
    limit: int = typer.Option(50),
    all: bool = typer.Option(False, "--all"),
    log_level: str = typer.Option("INFO"),
):
    _setup_logger(log_level)

    config_path = resolve_profile_config(profile)

    expanded: list[Path] = []
    for p in paths:
        if p.is_dir():
            expanded.extend(sorted(p.iterdir()))
        else:
            expanded.append(p)

    if not all and len(expanded) > limit:
        typer.echo(f"Showing first {limit} files")
        expanded = expanded[:limit]

    results = explain_files(str(config_path), expanded, verbose=verbose)

    for r in results:
        logger.info(format_explain(r, fmt=fmt, verbose=verbose))


@app.command()
def validate(
    profile: str = typer.Option(..., "--profile", "-p"),
    log_level: str = typer.Option("INFO"),
):
    _setup_logger(log_level)

    config_path = resolve_profile_config(profile)
    result = validate_config_file(Path(config_path))

    if result.ok:
        logger.info("Config is valid ✅")
        raise typer.Exit(code=0)

    logger.error("Config is invalid ❌")
    for err in result.errors:
        logger.error(f"- {err}")

    raise typer.Exit(code=1)


@profile_app.command("init")
def profile_init(
    name: str = typer.Argument(...),
    force: bool = typer.Option(False, "--force"),
):
    """
    Create profile folder + rules.yaml and register in profiles.yaml
    """

    base = _home()
    rules_dir = base / "rules" / name
    rules_dir.mkdir(parents=True, exist_ok=True)

    rules_yaml = rules_dir / "rules.yaml"

    if rules_yaml.exists() and not force:
        typer.echo("rules.yaml already exists. Use --force.")
        raise typer.Exit(1)

    rules_yaml.write_text(DEFAULT_CONFIG, encoding="utf-8")

    profiles_file = _profiles_file()
    profiles_file.parent.mkdir(parents=True, exist_ok=True)

    if profiles_file.exists():
        data = yaml.safe_load(profiles_file.read_text()) or {}
    else:
        data = {}

    data.setdefault("profiles", {})[name] = str(rules_dir)

    profiles_file.write_text(yaml.safe_dump(data, sort_keys=False))

    typer.echo(f"Profile '{name}' created.")
    typer.echo(f"Rules: {rules_yaml}")
    typer.echo(f"Registered in {profiles_file}")


@profile_app.command("list")
def profile_list(
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Show main settings (source/default_target/conflict_policy)"),
):
    """
    List all profiles
    """
    pf = _profiles_file()

    if not pf.exists():
        typer.echo("profiles.yaml not found. Run: filemindr profile init <name>")
        raise typer.Exit(1)

    data = yaml.safe_load(pf.read_text(encoding="utf-8")) or {}
    profiles = data.get("profiles", {})

    if not profiles:
        typer.echo("No profiles defined.")
        raise typer.Exit()

    for name in sorted(profiles.keys()):
        rules_dir = Path(profiles[name]).expanduser()
        rules_yaml = rules_dir / "rules.yaml"

        if not verbose:
            typer.echo(f"{name} -> {rules_dir}")
            continue

        if not rules_yaml.exists():
            typer.echo(f"{name} -> {rules_dir}  (rules.yaml missing)")
            continue

        cfg = yaml.safe_load(rules_yaml.read_text(encoding="utf-8")) or {}
        typer.echo(f"{name} -> {rules_dir}")
        typer.echo(f"  source:         {cfg.get('source')}")
        typer.echo(f"  default_target: {cfg.get('default_target')}")
        typer.echo(f"  conflict_policy:{cfg.get('conflict_policy')}")


@profile_app.command("edit")
def profile_edit(name: str = typer.Argument(...)):
    """
    Open profile rules.yaml in an editor (VISUAL/EDITOR or OS default fallback).
    """
    rules_yaml = Path(resolve_profile_config(name))
    open_in_editor(rules_yaml)


@profile_app.command("open")
def profile_open(name: str = typer.Argument(...)):
    """
    Open profile rules.yaml with the OS default application.
    """
    rules_yaml = Path(resolve_profile_config(name))
    open_with_default_app(rules_yaml)


@profile_app.command("show")
def profile_show(name: str = typer.Argument(...)):
    """
    Show resolved paths and main settings of a profile.
    """
    rules_yaml = Path(resolve_profile_config(name))
    cfg = yaml.safe_load(rules_yaml.read_text(encoding="utf-8")) or {}

    typer.echo(f"profile:         {name}")
    typer.echo(f"profile_dir:     {rules_yaml.parent}")
    typer.echo(f"rules.yaml:      {rules_yaml}")
    typer.echo("")
    typer.echo(f"source:          {cfg.get('source')}")
    typer.echo(f"default_target:  {cfg.get('default_target')}")
    typer.echo(f"conflict_policy: {cfg.get('conflict_policy')}")


@profile_app.command("remove")
def profile_remove(
    name: str = typer.Argument(...),
    yes: bool = typer.Option(False, "--yes", "-y", help="Skip confirmation"),
):
    """
    Completely remove profile: registry entry + folder (rules.yaml, etc).
    """
    pf = _profiles_file()
    if not pf.exists():
        typer.echo("profiles.yaml not found. Run: filemindr profile init <name>")
        raise typer.Exit(1)

    data = yaml.safe_load(pf.read_text(encoding="utf-8")) or {}
    profiles = data.get("profiles", {})

    if name not in profiles:
        typer.echo(f"Profile '{name}' not found.\n\nCreate it with:\n  filemindr profile init {name}")
        raise typer.Exit(1)

    profile_dir = Path(profiles[name]).expanduser()

    if not yes:
        if not typer.confirm(f"Delete profile '{name}' and ALL its files at:\n{profile_dir}\nContinue?"):
            raise typer.Exit()

    del profiles[name]
    pf.write_text(yaml.safe_dump(data, sort_keys=False), encoding="utf-8")

    if profile_dir.exists():
        import shutil
        shutil.rmtree(profile_dir)

    typer.echo(f"Profile '{name}' completely removed.")



@app.command()
def doctor():
    """
    Sanity checks on all profiles:
    - profiles.yaml exists
    - profile directory exists
    - rules.yaml exists
    - YAML validates
    - source/default_target paths resolve and exist (warn/create suggestion)
    """
    pf = _profiles_file()

    if not pf.exists():
        typer.echo("profiles.yaml not found. Run: filemindr profile init <name>")
        raise typer.Exit(1)

    data = yaml.safe_load(pf.read_text(encoding="utf-8")) or {}
    profiles = data.get("profiles", {})

    if not profiles:
        typer.echo("No profiles defined.")
        raise typer.Exit()

    ok = True

    for name in sorted(profiles.keys()):
        rules_dir = Path(profiles[name]).expanduser()
        rules_yaml = rules_dir / "rules.yaml"

        if not rules_dir.exists():
            typer.echo(f"❌ {name}: profile directory not found -> {rules_dir}")
            ok = False
            continue

        if not rules_yaml.exists():
            typer.echo(f"❌ {name}: rules.yaml missing -> {rules_yaml}")
            ok = False
            continue

        result = validate_config_file(rules_yaml)
        if not result.ok:
            typer.echo(f"❌ {name}: invalid config")
            for err in result.errors:
                typer.echo(f"   - {err}")
            ok = False
            continue

        cfg = yaml.safe_load(rules_yaml.read_text(encoding="utf-8")) or {}
        source = cfg.get("source")
        default_target = cfg.get("default_target")

        source_path = _expand(source) if isinstance(source, str) else None
        target_path = _expand(default_target) if isinstance(default_target, str) else None

        warn = False

        if source_path and not source_path.exists():
            typer.echo(f"⚠ {name}: source does not exist -> {source_path}")
            warn = True

        if target_path and not target_path.exists():
            typer.echo(f"⚠ {name}: default_target does not exist -> {target_path}")
            warn = True

        if warn:
            typer.echo(f"✔ {name}: config valid (with warnings)")
        else:
            typer.echo(f"✔ {name}: OK")

    if not ok:
        raise typer.Exit(1)


if __name__ == "__main__":
    app()



from __future__ import annotations

import signal
import time
from dataclasses import dataclass, field
from pathlib import Path
from queue import Queue, Empty
from threading import Event, Thread

from loguru import logger
from watchfiles import watch

from filemindr.core.runner import run_pipeline


DEFAULT_IGNORED_SUFFIXES = {
    ".crdownload",
    ".part",
    ".aria2",
    ".tmp",
    ".temp",
}

DEFAULT_IGNORED_PREFIXES = {"~$"}


@dataclass(frozen=True)
class WatchOptions:
    debounce_ms: int = 800
    stable_ms: int = 1500
    poll_interval_ms: int = 250

    ignore_suffixes: set[str] = field(default_factory=lambda: set(DEFAULT_IGNORED_SUFFIXES))
    ignore_prefixes: set[str] = field(default_factory=lambda: set(DEFAULT_IGNORED_PREFIXES))

    max_pending_paths: int = 10_000
    log_pending: bool = False

    main_loop_tick_ms: int = 250


def _should_ignore(path: Path, opts: WatchOptions) -> bool:
    name = path.name.lower()
    if path.name.lower() == "filemindr.yaml":
        return True
    for p in opts.ignore_prefixes:
        if name.startswith(p.lower()):
            return True
    for s in opts.ignore_suffixes:
        if name.endswith(s.lower()):
            return True
    return False


def _is_stable(path: Path, stable_ms: int, poll_interval_ms: int) -> bool:
    if not path.exists() or not path.is_file():
        return False

    stable_s = stable_ms / 1000
    poll_s = max(0.05, poll_interval_ms / 1000)

    start = time.monotonic()
    try:
        st0 = path.stat()
        prev_size = st0.st_size
        prev_mtime = st0.st_mtime
    except OSError:
        return False

    while True:
        time.sleep(poll_s)

        if not path.exists():
            return False

        try:
            st = path.stat()
            size = st.st_size
            mtime = st.st_mtime
        except OSError:
            return False

        if size != prev_size or mtime != prev_mtime:
            prev_size, prev_mtime = size, mtime
            start = time.monotonic()

        if (time.monotonic() - start) >= stable_s:
            return True


def watch_and_run(
    source_dir: Path,
    config_path: str,
    dry_run: bool,
    opts: WatchOptions | None = None,
    *,
    once: bool = False,
) -> None:
    opts = opts or WatchOptions()
    source_dir = source_dir.expanduser().resolve()

    logger.info(f"Watching: {source_dir}")
    logger.info(f"Config: {config_path} | dry_run={dry_run}")
    logger.info(f"debounce_ms={opts.debounce_ms} stable_ms={opts.stable_ms} once={once}")

    q: Queue[Path] = Queue()
    stop_event = Event()

    prev_handler = signal.getsignal(signal.SIGINT)

    def _sigint_handler(signum, frame):
        if not stop_event.is_set():
            logger.info("Ctrl+C received. Stopping watcher...")
            stop_event.set()

    signal.signal(signal.SIGINT, _sigint_handler)

    def producer() -> None:
        try:
            for changes in watch(source_dir, recursive=False, stop_event=stop_event):
                if stop_event.is_set():
                    break
                for _, changed_path in changes:
                    p = Path(changed_path)
                    if p.is_dir():
                        continue
                    if _should_ignore(p, opts):
                        continue
                    q.put(p)
        except Exception:
            logger.exception("Watcher producer crashed.")
            stop_event.set()

    producer_thread = Thread(target=producer, daemon=False)
    producer_thread.start()

    pending: set[Path] = set()

    def flush_pending() -> None:
        nonlocal pending
        if not pending:
            return

        if opts.log_pending:
            logger.debug(
                f"Pending paths ({len(pending)}): " + ", ".join(str(p) for p in list(pending)[:50])
            )

        ready = {
            p for p in pending
            if not _should_ignore(p, opts)
            and _is_stable(p, opts.stable_ms, opts.poll_interval_ms)
        }

        if ready:
            logger.info(f"Detected stable changes ({len(ready)} paths). Running pipeline...")
            try:
                run_pipeline(config_path, dry_run=dry_run, only_paths=ready)
            except KeyboardInterrupt:
                stop_event.set()
                logger.info("Pipeline interrupted by user.")
            except Exception:
                logger.exception("Pipeline run failed.")
            finally:
                if once and not stop_event.is_set():
                    logger.info("watch --once completed. Exiting watcher.")
                    stop_event.set()

        pending.clear()

    quiet_s = opts.debounce_ms / 1000
    tick_s = max(0.05, opts.main_loop_tick_ms / 1000)

    try:
        while not stop_event.is_set():
            try:
                p = q.get(timeout=tick_s)
            except Empty:
                continue

            pending.add(p)

            if len(pending) > opts.max_pending_paths:
                logger.warning(f"Too many pending paths ({len(pending)}). Flushing now...")
                flush_pending()
                continue

            deadline = time.monotonic() + quiet_s
            while not stop_event.is_set():
                timeout = deadline - time.monotonic()
                if timeout <= 0:
                    break
                try:
                    p2 = q.get(timeout=min(timeout, tick_s))
                except Empty:
                    continue
                else:
                    pending.add(p2)
                    deadline = time.monotonic() + quiet_s

            if stop_event.is_set():
                break

            flush_pending()

    except KeyboardInterrupt:
        stop_event.set()
        logger.info("Watcher stopped by user (KeyboardInterrupt).")
    finally:
        if pending and not stop_event.is_set():
            flush_pending()

        stop_event.set()
        producer_thread.join(timeout=2.0)

        signal.signal(signal.SIGINT, prev_handler)

        logger.info("Watcher shutdown complete.")



from __future__ import annotations


def test_doctor_ok(runner, fake_home, cli_app, tmp_path):
    r1 = runner.invoke(cli_app, ["profile", "init", "home"])
    assert r1.exit_code == 0

    src = tmp_path / "src"
    src.mkdir(parents=True, exist_ok=True)

    rules_file = fake_home / ".filemindr" / "rules" / "home" / "rules.yaml"
    rules_file.write_text(
        f"""
source: {src.as_posix()}
default_target: {src.as_posix()}/others
conflict_policy: rename

rules:
  - name: images
    priority: 10
    match:
      extensions: ["jpg"]
    action:
      move_to: {src.as_posix()}/images
""".strip(),
        encoding="utf-8",
    )

    result = runner.invoke(cli_app, ["doctor"])
    assert result.exit_code == 0, (result.stdout or "") + (result.stderr or "")

def test_doctor_fails_when_rules_missing(runner, fake_home, cli_app):
    runner.invoke(cli_app, ["profile", "init", "home"])

    rules_file = fake_home / ".filemindr" / "rules" / "home" / "rules.yaml"
    rules_file.unlink()

    result = runner.invoke(cli_app, ["doctor"])
    assert result.exit_code != 0



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

import filemindr.core.runner as runner


def test_trash_calls_send2trash(monkeypatch, tmp_path: Path):
    target = tmp_path / "existing.txt"
    target.write_text("old", encoding="utf-8")

    calls = {"n": 0, "arg": None}

    def fake_send2trash(arg: str):
        calls["n"] += 1
        calls["arg"] = arg

    monkeypatch.setitem(
        runner.__dict__,
        "send2trash",
        SimpleNamespace(send2trash=fake_send2trash),
    )

    import sys
    sys.modules["send2trash"] = SimpleNamespace(send2trash=fake_send2trash)

    runner._trash(target)

    assert calls["n"] == 1
    assert calls["arg"] == str(target)


def test_trash_falls_back_to_unlink_when_send2trash_fails(monkeypatch, tmp_path: Path):
    target = tmp_path / "existing.txt"
    target.write_text("old", encoding="utf-8")

    def failing_send2trash(_: str):
        raise RuntimeError("boom")

    import sys
    sys.modules["send2trash"] = SimpleNamespace(send2trash=failing_send2trash)

    unlinked = {"n": 0}

    real_unlink = Path.unlink

    def fake_unlink(self, missing_ok: bool = False):
        unlinked["n"] += 1
        return real_unlink(self, missing_ok=missing_ok)

    monkeypatch.setattr(Path, "unlink", fake_unlink, raising=True)

    runner._trash(target)

    assert unlinked["n"] == 1
    assert not target.exists()

def test_policy_trash_replaces_existing_destination(monkeypatch, tmp_path: Path):
    """
    Integração leve: simula conflito e garante que:
    - o destino existente é "trashed"
    - o novo arquivo é copiado/movido para o destino final
    """
    source_dir = tmp_path / "Downloads"
    source_dir.mkdir()

    dest_dir = tmp_path / "images"
    dest_dir.mkdir()

    src = source_dir / "pic.jpg"
    src.write_text("new", encoding="utf-8")

    dest = dest_dir / "pic.jpg"
    dest.write_text("old", encoding="utf-8")

    cfg = tmp_path / "filemindr.yaml"
    cfg.write_text(
        f"""
source: {source_dir}
default_target: {tmp_path / "others"}
conflict_policy: trash
rules:
  - name: images
    priority: 10
    match:
      extensions: ["jpg"]
    action:
      copy_to: {dest_dir}
""".strip(),
        encoding="utf-8",
    )

    trashed = {"n": 0, "arg": None}

    def fake_trash(p: Path):
        trashed["n"] += 1
        trashed["arg"] = p
        if p.exists():
            p.unlink()

    monkeypatch.setattr(runner, "_trash", fake_trash)

    runner.run_pipeline(str(cfg), dry_run=False)

    assert trashed["n"] == 1
    assert trashed["arg"] == dest
    assert dest.exists()
    assert dest.read_text(encoding="utf-8") == "new"


from __future__ import annotations

import time
from pathlib import Path


def test_watch_normal_runs_twice_then_stops(monkeypatch, tmp_path: Path):
    """
    Normal watch (once=False): runs pipeline for each burst (debounce window).
    We simulate 2 bursts separated by > debounce, then stop.
    """
    import filemindr.core.watcher as watcher

    source_dir = tmp_path / "Downloads"
    source_dir.mkdir()

    cfg_path = tmp_path / "filemindr.yaml"
    cfg_path.write_text("source: .\n", encoding="utf-8")

    f1 = source_dir / "a.jpg"
    f2 = source_dir / "b.jpg"
    f1.write_text("x", encoding="utf-8")
    f2.write_text("y", encoding="utf-8")

    monkeypatch.setattr(watcher, "_is_stable", lambda *args, **kwargs: True)
    monkeypatch.setattr(watcher, "_should_ignore", lambda *args, **kwargs: False)

    calls: list[set[Path] | None] = []

    def fake_run_pipeline(config_path: str, dry_run: bool = False, only_paths=None):
        calls.append(only_paths)

    monkeypatch.setattr(watcher, "run_pipeline", fake_run_pipeline)

    opts = watcher.WatchOptions(debounce_ms=10, stable_ms=10, poll_interval_ms=10)

    def fake_watch(_dir, recursive=False, stop_event=None):
        yield [(1, str(f1))]
        time.sleep(0.05)

        yield [(1, str(f2))]
        time.sleep(0.05)

        if stop_event is not None:
            stop_event.set()

    monkeypatch.setattr(watcher, "watch", fake_watch)

    watcher.watch_and_run(
        source_dir=source_dir,
        config_path=str(cfg_path),
        dry_run=False,
        opts=opts,
        once=False,
    )

    assert len(calls) == 2
    assert calls[0] is not None and f1 in calls[0]
    assert calls[1] is not None and f2 in calls[1]


def test_watch_once_runs_once_and_exits(monkeypatch, tmp_path: Path):
    """
    Watch once (once=True): should run pipeline only once and exit
    after the first stable burst.
    """
    import filemindr.core.watcher as watcher

    source_dir = tmp_path / "Downloads"
    source_dir.mkdir()

    cfg_path = tmp_path / "filemindr.yaml"
    cfg_path.write_text("source: .\n", encoding="utf-8")

    f1 = source_dir / "pic.jpg"
    f1.write_text("x", encoding="utf-8")

    monkeypatch.setattr(watcher, "_is_stable", lambda *args, **kwargs: True)
    monkeypatch.setattr(watcher, "_should_ignore", lambda *args, **kwargs: False)

    calls: list[set[Path] | None] = []

    def fake_run_pipeline(config_path: str, dry_run: bool = False, only_paths=None):
        calls.append(only_paths)

    monkeypatch.setattr(watcher, "run_pipeline", fake_run_pipeline)

    opts = watcher.WatchOptions(debounce_ms=10, stable_ms=10, poll_interval_ms=10)

    def fake_watch(_dir, recursive=False, stop_event=None):
        yield [(1, str(f1))]
        yield [(1, str(f1))]

    monkeypatch.setattr(watcher, "watch", fake_watch)

    watcher.watch_and_run(
        source_dir=source_dir,
        config_path=str(cfg_path),
        dry_run=False,
        opts=opts,
        once=True,
    )

    assert len(calls) == 1
    assert calls[0] is not None and f1 in calls[0]



import os
import time
from pathlib import Path

from filemindr.core.runner import _load_rules, _match_rule


def test_older_than_days(tmp_path: Path):
    config = {
        "rules": [
            {
                "name": "old_installers",
                "priority": 50,
                "match": {"extensions": ["exe"], "older_than_days": 1},
                "action": {"move_to": str(tmp_path / "old")},
            }
        ]
    }
    rules = _load_rules(config)

    f = tmp_path / "setup.exe"
    f.write_text("x")

    two_days_ago = time.time() - (2 * 24 * 60 * 60)
    os.utime(f, (two_days_ago, two_days_ago))

    rule = _match_rule(f, rules)
    assert rule is not None
    assert rule.name == "old_installers"


from __future__ import annotations


def test_profile_init_creates_rules_and_registers(runner, fake_home, cli_app):
    result = runner.invoke(cli_app, ["profile", "init", "home"])
    assert result.exit_code == 0

    base = fake_home / ".filemindr"
    rules_file = base / "rules" / "home" / "rules.yaml"
    profiles_file = base / "profiles.yaml"

    assert rules_file.exists()
    assert profiles_file.exists()

    text = profiles_file.read_text(encoding="utf-8")
    assert "home" in text



import os
import sys
import shlex
import subprocess
from pathlib import Path


def _editor_cmd() -> list[str]:
    ed = os.environ.get("VISUAL") or os.environ.get("EDITOR")
    if ed:
        return shlex.split(ed)

    if sys.platform.startswith("win"):
        return ["notepad.exe"]
    if sys.platform == "darwin":
        return ["open", "-e"]
    return ["nano"]


def open_in_editor(path: Path) -> None:
    cmd = _editor_cmd() + [str(path)]
    subprocess.run(cmd, check=False)


def open_with_default_app(path: Path) -> None:
    if sys.platform.startswith("win"):
        os.startfile(str(path))  # type: ignore[attr-defined]
        return
    if sys.platform == "darwin":
        subprocess.run(["open", str(path)], check=False)
        return
    subprocess.run(["xdg-open", str(path)], check=False)
